#define _GNU_SOURCE
#include <assert.h>
#include <err.h>
#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <unistd.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <stdbool.h>
#include <sched.h>
#include <sys/stat.h>
#include <sys/mount.h>
#include <sys/prctl.h>
#include <sys/socket.h>
#include <sys/uio.h>
#include <sys/mman.h>
#include <linux/fuse.h>

#define SYSCHK(x) ({          \
  typeof(x) __res = (x);      \
  if (__res == (typeof(x))-1) \
    err(1, "SYSCHK(" #x ")"); \
  __res;                      \
})

static void hexdump(void *_data, size_t byte_count) {
  printf("hexdump(%p, 0x%lx)\n", _data, (unsigned long)byte_count);
  for (unsigned long byte_offset = 0; byte_offset < byte_count; byte_offset += 16) {
    unsigned char *bytes = ((unsigned char*)_data) + byte_offset;
    unsigned long line_bytes = (byte_count - byte_offset > 16) ?
            16 : (byte_count - byte_offset);
    char line[1000];
    char *linep = line;
    linep += sprintf(linep, "%08lx  ", byte_offset);
    for (int i=0; i<16; i++) {
      if (i >= line_bytes) {
        linep += sprintf(linep, "   ");
      } else {
        linep += sprintf(linep, "%02hhx ", bytes[i]);
      }
    }
    linep += sprintf(linep, " |");
    for (int i=0; i<line_bytes; i++) {
      if (isalnum(bytes[i]) || ispunct(bytes[i]) || bytes[i] == ' ') {
        *(linep++) = bytes[i];
      } else {
        *(linep++) = '.';
      }
    }
    linep += sprintf(linep, "|");
    puts(line);
  }
}

static void write_file(char *name, char *buf) {
  int fd = SYSCHK(open(name, O_WRONLY));
  if (write(fd, buf, strlen(buf)) != strlen(buf))
    err(1, "write %s", name);
  close(fd);
}

static void write_map(char *name, int outer_id) {
  char buf[100];
  sprintf(buf, "0 %d 1", outer_id);
  write_file(name, buf);
}

#define NAME1 "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB"
#define NAME2 "BBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBBB2"
#define PATH1 "/tmp/"NAME1
#define PATH2 "/tmp/"NAME2

static int fuse_fd;
#define READ(_obj) if (read(fuse_fd, &(_obj), sizeof(_obj)) != sizeof(_obj)) err(1, "failed to read " #_obj)
#define WRITE(_obj) if (write(fuse_fd, &(_obj), (_obj).h.len) != (_obj).h.len) err(1, "failed to write " #_obj)
static void fuse_server(void) {
  struct fuse_attr_out dir_attrs = { .attr_valid = FATTR_SIZE | FATTR_MODE, .attr = { .size = 42, .mode = S_IFDIR | 0777 } };
  while (1) {
    struct inbuf {
      struct fuse_in_header inh;
      union {
        struct fuse_init_out init_out;
        struct fuse_open_in open_in;
        struct fuse_read_in read_in;
        char pad[10000];
      };
    } buf;
    size_t read_len = SYSCHK(read(fuse_fd, &buf, sizeof(buf)));
    assert(read_len >= sizeof(buf.inh));
    if (buf.inh.opcode == FUSE_INIT) {
      printf("fuse: init\n");
      struct {
        struct fuse_out_header h;
        struct fuse_init_in b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = { .major = buf.init_out.major, .minor = buf.init_out.minor }
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_LOOKUP) {
      printf("fuse: lookup, node=%u, name hexdump:\n", buf.inh.nodeid);
      hexdump(buf.pad, read_len - offsetof(struct inbuf, pad));
      if (buf.inh.nodeid == 1 && strcmp(buf.pad, NAME1) == 0) {
        struct {
          struct fuse_out_header h;
          struct fuse_entry_out b;
        } reply = {
          .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
          .b = {
            .nodeid = 2,
            .entry_valid = 1,
            .attr = dir_attrs.attr
          }
        };
        WRITE(reply);
      } else {
        struct {
          struct fuse_out_header h;
        } reply = {
          .h = { .len = sizeof(reply), .error = -ENOENT, .unique = buf.inh.unique },
        };
        WRITE(reply);
      }
    } else if (buf.inh.opcode == FUSE_GETATTR) {
      printf("fuse: getattr, node=%u\n", buf.inh.nodeid);
      struct {
        struct fuse_out_header h;
        struct fuse_attr_out b;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
        .b = dir_attrs
      };
      WRITE(reply);
    } else if (buf.inh.opcode == FUSE_RENAME) {
      printf("fuse: rename, BEGIN DELAY\n");
      sleep(3);
      printf("fuse: rename, END DELAY\n");
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = 0, .unique = buf.inh.unique },
      };
      WRITE(reply);
      printf("fuse: replied to rename, BEGIN MORE DELAY\n");
      sleep(30);
      printf("fuse: resuming processing\n");
    } else {
      printf("FUSE_??? %d !!!!!!!!!!!!!\n", buf.inh.opcode);
      struct {
        struct fuse_out_header h;
      } reply = {
        .h = { .len = sizeof(reply), .error = -ENOSYS, .unique = buf.inh.unique },
      };
      WRITE(reply);
    }
  }
  close(fuse_fd);
}

int main(void) {
  // boilerplate start code
  setbuf(stdout, NULL);
  setbuf(stderr, NULL);

  // set up user namespace (for unprivileged access to FUSE)
  int outer_uid = getuid();
  int outer_gid = getgid();
  SYSCHK(unshare(CLONE_NEWUSER|CLONE_NEWNS));
  SYSCHK(mount(NULL, "/", NULL, MS_PRIVATE|MS_REC, NULL));
  write_file("/proc/self/setgroups", "deny");
  write_map("/proc/self/uid_map", outer_uid);
  write_map("/proc/self/gid_map", outer_gid);

  // create a FUSE mount, and handle requests in a child
  fuse_fd = SYSCHK(open("/dev/fuse", O_RDWR));
  char mount_data[4096];
  sprintf(mount_data, "fd=%d,rootmode=0040777,user_id=%d,group_id=%d", fuse_fd, getuid(), getgid());
  SYSCHK(mount("blah", "/tmp", "fuse", MS_NODEV|MS_NOSUID, mount_data));
  if (SYSCHK(fork()) == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    fuse_server();
    exit(0);
  }

  prctl(PR_SET_NAME, "TRACEME");

  SYSCHK(access(PATH1, F_OK));
  pid_t rename_child = SYSCHK(fork());
  if (rename_child == 0) {
    SYSCHK(prctl(PR_SET_PDEATHSIG, SIGKILL));
    printf("rename_child: begin rename\n");
    SYSCHK(rename(PATH1, PATH2));
    printf("rename_child: done with rename\n");
    exit(0);
  }
  sleep(2);
  printf("begin revalidating access\n");
  SYSCHK(access(PATH1"/foo", F_OK));
}
